from typing import Any, Awaitable, Callable

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import HumanMessage
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import RunnableConfig
from observe.logger import get_logger

from llm.common import PredictionError

logger = get_logger()


def ask_again_convo(
    model: BaseChatModel,
    model_config: RunnableConfig,
    prompt: ChatPromptTemplate,
    prompt_params: dict[str, Any],
    prompt_ask_again_message: HumanMessage,
    format_response_fn: Callable[[str], Any],
    validate_response_fn: Callable[[Any], bool],
    retry: int = 4,
    label: str = None,
) -> Any:
    # ask question the first time
    logger.info("Asking question the first time", label=label)
    chain = prompt | model
    res_ans_first = chain.invoke(prompt_params, model_config)
    ans_first = format_response_fn(res_ans_first.content)

    # validate if the answer is correct
    if validate_response_fn(ans_first):
        return ans_first
    logger.info(
        "Answer is not correct. Asking question again", label=label, ans_first=ans_first
    )

    # start asking question again loop
    iteration = 0
    prompt_ask_again = ChatPromptTemplate.from_messages(
        [
            MessagesPlaceholder(variable_name="chat_history"),
            prompt_ask_again_message,  # first ask again message
        ]
    )
    chat_ask_again_history = []
    while iteration < retry:
        chat_history = (
            prompt.format_messages(**prompt_params)
            + [res_ans_first]
            + chat_ask_again_history
        )
        prompt_ask_again_params = {"chat_history": chat_history}
        chain_ask_again = prompt_ask_again | model
        # ask question again
        res_ans_again = chain_ask_again.invoke(prompt_ask_again_params, model_config)
        ans_again = format_response_fn(res_ans_again.content)
        if validate_response_fn(ans_again):
            return ans_again

        iteration += 1
        chat_ask_again_history.extend([prompt_ask_again_message, res_ans_again])
        logger.info(
            "Answer is not correct. Asking question again",
            label=label,
            retry=iteration,
            ans_again=ans_again,
        )

    raise PredictionError("Unable to predict a valid answer")


async def async_ask_again_convo(
    model: BaseChatModel,
    model_config: RunnableConfig,
    prompt: ChatPromptTemplate,
    prompt_params: dict[str, Any],
    prompt_ask_again_message: HumanMessage,
    format_response_fn: Callable[[str], Awaitable[Any]],
    validate_response_fn: Callable[[Any], Awaitable[bool]],
    retry: int = 4,
    label: str = None,
) -> Any:
    # ask question the first time
    logger.info("Asking question the first time", label=label)
    chain = prompt | model
    res_ans_first = await chain.ainvoke(prompt_params, model_config)
    ans_first = await format_response_fn(res_ans_first.content)

    # validate if the answer is correct
    if await validate_response_fn(ans_first):
        return ans_first
    logger.info(
        "Answer is not correct. Asking question again", label=label, ans_first=ans_first
    )

    # start asking question again loop
    iteration = 0
    prompt_ask_again = ChatPromptTemplate.from_messages(
        [
            MessagesPlaceholder(variable_name="chat_history"),
            prompt_ask_again_message,  # first ask again message
        ]
    )
    chat_ask_again_history = []
    while iteration < retry:
        chat_history = (
            prompt.format_messages(**prompt_params)
            + [res_ans_first]
            + chat_ask_again_history
        )
        prompt_ask_again_params = {"chat_history": chat_history}
        chain_ask_again = prompt_ask_again | model
        # ask question again
        res_ans_again = await chain_ask_again.ainvoke(
            prompt_ask_again_params, model_config
        )
        ans_again = await format_response_fn(res_ans_again.content)
        if await validate_response_fn(ans_again):
            return ans_again

        iteration += 1
        chat_ask_again_history.extend([prompt_ask_again_message, res_ans_again])
        logger.info(
            "Answer is not correct. Asking question again",
            label=label,
            retry=iteration,
            ans_again=ans_again,
        )

    raise PredictionError("Unable to predict a valid answer")



import base64
import hashlib
import json
from typing import List

from langchain_core.embeddings.embeddings import Embeddings
from langchain_openai import OpenAIEmbeddings
from redis import Redis


def hash_string_base64(s, max_length=100):
    # Hash the string using SHA-256
    hash_object = hashlib.sha256(s.encode())
    hash_bytes = hash_object.digest()  # Get the hash as bytes

    # Encode the hash in base64
    hash_base64 = base64.b64encode(hash_bytes).decode("utf-8")

    # Truncate or somehow reduce the size to meet max_length requirement
    # Note: You might want to adjust the logic to ensure the final string meets your exact requirements
    truncated_hash = hash_base64[:max_length]

    return truncated_hash


class RedisCachedOpenAIEmbeddings(Embeddings):
    def __init__(self, redis_client: Redis):
        self.redis_client = redis_client
        self.openai_client = OpenAIEmbeddings()

    @staticmethod
    def _get_key(text):
        return f"llm-embeddings-openai-{hash_string_base64(text)}"

    def _get_from_cache(self, text) -> list[float] | None:
        value = self.redis_client.get(self._get_key(text))
        if value is None:
            return None
        return json.loads(value)

    def _set_in_cache(self, text, value):
        self.redis_client.set(self._get_key(text), json.dumps(value))

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        # get cached values
        cached_documents = {text: self._get_from_cache(text) for text in texts}
        # embed new values
        uncached_texts = [text for text in texts if cached_documents.get(text) is None]
        documents = self.openai_client.embed_documents(uncached_texts)
        # add to cache
        new_documents = {}
        for text, document in zip(uncached_texts, documents):
            self._set_in_cache(text, document)
            new_documents[text] = document

        # return all values in the same order
        return [
            cached_documents.get(text)
            if cached_documents.get(text) is not None
            else new_documents.get(text)
            for text in texts
        ]

    def embed_query(self, text: str) -> List[float]:
        # get cached value
        cached_value = self._get_from_cache(text)
        if cached_value is not None:
            return cached_value
        # embed new value
        document = self.openai_client.embed_query(text)
        # add to cache
        self._set_in_cache(text, document)
        return document



from typing import Any, Awaitable, Callable

from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import RunnableConfig
from observe.logger import get_logger

from llm.common import InternalPredictionError, PredictionError

logger = get_logger()


async def retry_convo(
    model: BaseChatModel,
    model_config: RunnableConfig,
    prompt: ChatPromptTemplate,
    prompt_params: dict[str, Any],
    format_response_fn: Callable[[str], Awaitable[Any]],
    validate_response_fn: Callable[[Any], Awaitable[bool]],
    retry: int = 4,
    label: str = None,
) -> Any:
    logger.info("Starting retry convo loop", label=label)
    iteration = 0
    while iteration < retry:
        # ask question
        logger.info("Asking question", label=label)
        chain = prompt | model
        res_ans = await chain.ainvoke(prompt_params, model_config)
        ans = await format_response_fn(res_ans.content)

        # validate if the answer is correct
        try:
            if await validate_response_fn(ans):
                return ans
            logger.error(
                "Answer is not correct. Asking question again", label=label, ans=ans
            )
        except InternalPredictionError as e:
            logger.error(
                "Answer is not correct. Asking question again",
                label=label,
                ans=ans,
                error=str(e),
            )

        iteration += 1

    raise PredictionError("reached max retry limit")



from typing import Any, Dict, Iterator, List, Optional
from urllib.parse import urljoin

import httpx
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage, HumanMessage, SystemMessage
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from openai.types.chat import ChatCompletion


class ProxyChatModel(BaseChatModel):
    api_key: str
    endpoint: str
    model_name: str = "proxy"
    timeout: int = 60 * 5

    # chat related
    # Definitions taken from https://platform.openai.com/docs/api-reference/chat/create?lang=python
    frequency_penalty: int = 0
    logit_bias: Any = None
    logprobs: bool | None = None
    top_logprobs: int | None = None
    max_tokens: int | None = None
    n: int = 1
    presence_penalty: int = 0
    response_format: str | None = None
    seed: int | None = None
    stop: str | list | None = None
    # stream: bool = False
    temperature: float = 0.0
    top_p: float = 1
    user: str | None = None

    @property
    def _llm_type(self) -> str:
        return f"proxy-server:{self.model_name}"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        return {
            "model_name": self.model_name,
        }

    @staticmethod
    def _format_message(message: BaseMessage) -> dict:
        if isinstance(message, SystemMessage):
            return {
                "content": message.content,
                "role": "system",
            }
        if isinstance(message, HumanMessage):
            return {
                "content": message.content,
                "role": "user",
            }
        if isinstance(message, AIMessage):
            return {
                "content": message.content,
                "role": "assistant",
            }
        raise ValueError(f"Unknown message type: {type(message)}")

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """
        Args:
            messages: the prompt composed of a list of messages.
            stop: a list of strings on which the model should stop generating.
                  If generation stops due to a stop token, the stop token itself
                  SHOULD BE INCLUDED as part of the output. This is not enforced
                  across models right now, but it's a good practice to follow since
                  it makes it much easier to parse the output of the model
                  downstream and understand why generation stopped.
            run_manager: A run manager with callbacks for the LLM.
        """
        with httpx.Client() as client:
            formatted_messages = [self._format_message(message) for message in messages]
            config = "{}"
            res = client.post(
                urljoin(
                    self.endpoint,
                    f"/openai/deployments/{self.model_name}/chat/completions?api-version={config}",
                ),
                headers={
                    "Authorization": f"Bearer {self.api_key}",
                },
                json={
                    **{
                        "model": self.model_name,
                        "messages": formatted_messages,
                        "meta": run_manager.metadata,
                        # chat related
                        "frequency_penalty": self.frequency_penalty,
                        "logit_bias": self.logit_bias,
                        "logprobs": self.logprobs,
                        "top_logprobs": self.top_logprobs,
                        "max_tokens": self.max_tokens,
                        "n": self.n,
                        "presence_penalty": self.presence_penalty,
                        "response_format": self.response_format,
                        "seed": self.seed,
                        "stop": self.stop,
                        "temperature": self.temperature,
                        "top_p": self.top_p,
                        "user": self.user,
                    },
                    **kwargs,
                },
                timeout=self.timeout,
            )
            if res.status_code != 200:
                raise Exception(f"Failed to generate chat: {res.json()}")

            # build response
            body = ChatCompletion(**res.json())
            generations = [
                ChatGeneration(
                    message=AIMessage(
                        content=choice.message.content,
                    )
                )
                for choice in body.choices
            ]
            return ChatResult(generations=generations, llm_output=None)

        raise Exception("Failed to generate chat")

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        raise NotImplementedError("Streaming is not supported for this model.")



import asyncio
import time
from typing import Any, Callable, Coroutine

from observe.logger import get_logger

logger = get_logger()


class InternalPredictionError(Exception):
    pass


class PredictionError(Exception):
    pass


def exponential_backoff(func, max_retries=2):
    delay = 2  # Initial delay is seconds
    for i in range(max_retries):
        try:
            return func()
        except PredictionError as e:
            raise e
        except Exception as e:
            logger.error(
                f"Attempt failed. Retrying in {delay} seconds...",
                error=str(e),
                i=i + 1,
                delay=delay,
            )
            time.sleep(delay)
            delay *= 2  # Double the delay each time we retry

    logger.error("All attempts failed")
    raise PredictionError("All attempts failed")


AsyncCallback = Callable[[], Coroutine[Any, Any, Any]]


async def async_exponential_backoff(
    func: AsyncCallback, max_retries=2, initial_delay=2
):
    delay = initial_delay  # Initial delay in seconds
    for i in range(max_retries):
        try:
            return await func()
        except PredictionError as e:
            raise e
        except Exception as e:
            logger.error(
                f"Attempt failed. Retrying in {delay} seconds...",
                error=str(e),
                i=i + 1,
                delay=delay,
            )
            await asyncio.sleep(delay)
            delay *= 2  # Double the delay each time we retry

    logger.error("All attempts failed")
    raise PredictionError("All attempts failed")
